<?php

namespace Drupal\content_insights_report_group\Controller;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Database\Database;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Routing\RouteMatchInterface;

/**
 * Defines a controller for the content_insights_report_group module.
 */
class ContentInsightsReportGroupController extends ControllerBase {

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected $moduleHandler;

  public function __construct(ConfigFactoryInterface $config_factory, EntityTypeManagerInterface $entity_type_manager, ModuleHandlerInterface $module_handler) {
    $this->configFactory = $config_factory;
    $this->entityTypeManager = $entity_type_manager;
    $this->moduleHandler = $module_handler;
  }

  /**
   * Provides the results for a group.
   *
   * @param \Drupal\Core\Routing\RouteMatchInterface $route_match
   *   The route match service.
   *
   * @return array
   *   A render array.
   */
  public function contentReport(RouteMatchInterface $route_match) {

    // Retrieve the 'group' parameter from the URL.
    $group_id = $route_match->getParameter('group');
    // Load the group entity using the group ID.
    $group = $this->entityTypeManager->getStorage('group')->load($group_id);
    $group_data = [];

    if ($group) {
      // Get the group id (id).
      $group_data['id'] = $group_id;
      // Get the group name (title).
      $group_data['title'] = $group->label();
    }

    $config = $this->configFactory->get('content_insights_report_group.settings');
    $show_summary = $config->get('show_summary');
    $show_workflow_moderation_state_report = $config->get('show_workflow_moderation_state_report');
    $show_created_by_and_date = $config->get('show_created_by_and_date');
    $show_insights_report = $config->get('show_insights_report');
    $number_of_months = $config->get('number_of_months');
    $percentage_totals = $config->get('percentage_total');
    $sort_by_field = $config->get('sort_by_field');
    $sort_by_direction = $config->get('sort_by_direction');
    $show_print = $config->get('show_print');
    $show_empty_results = $config->get('show_empty_results');

    $report_data = [];

    // Fetch all content types.
    $content_types = $this->entityTypeManager->getStorage('node_type')->loadMultiple();

    // Fetch Workflow content_moderation Template.
    $moderation_state = [];
    if ($this->moduleHandler->moduleExists('content_moderation')) {
      $workflows = $this->entityTypeManager->getStorage('workflow')->loadMultiple();
      // Loop through each workflow and get its states.
      foreach ($workflows as $workflow) {
        // Get the states defined for this workflow.
        $workflow_states = $workflow->getTypePlugin()->getStates();
        foreach ($workflow_states as $state) {
          // Published Moderation state.
          $moderation_state[$workflow->id()]['name'] = $workflow->label();
          $moderation_state[$workflow->id()]['total_count'] = 0;
          $moderation_state[$workflow->id()]['status']['1']['name'] = 'Published';
          $moderation_state[$workflow->id()]['status']['1']['stage'][$state->id()]['name'] = $state->label();
          $moderation_state[$workflow->id()]['status']['1']['stage'][$state->id()]['count'] = 0;
          // Unpublished Moderation state.
          $moderation_state[$workflow->id()]['status']['0']['name'] = 'Unpublished';
          $moderation_state[$workflow->id()]['status']['0']['stage'][$state->id()]['name'] = $state->label();
          $moderation_state[$workflow->id()]['status']['0']['stage'][$state->id()]['count'] = 0;
        }
      }
    }

    switch ($percentage_totals) {
      case "published_nodes":
        // Sum of all published nodes.
        $percentage_totals_text = $this->t("Total percentage based on published nodes");
        break;

      case "all_nodes":
        // Sum of all nodes.
        $percentage_totals_text = $this->t("Total percentage based on all nodes");
        break;

      case "unpublished_nodes":
        // Sum of all published nodes.
        $percentage_totals_text = $this->t("Total percentage based on unpublished nodes");
        break;

    }

    // Prepare time intervals for # months.
    $current_date = new DrupalDateTime();
    $months_ago = (new DrupalDateTime("first day of this month"))->modify('-' . $number_of_months . ' months');

    foreach ($content_types as $content_type) {
      $type_id = $content_type->id();
      $type = $content_type->getOriginalId();
      $type_name = $content_type->label();
      $type_description = $content_type->get('description');
      $content_type_data = [];

      // Content type stats: Published and Unpublished nodes count.
      $published_nodes = $this->getNodeCount($content_type->id(), TRUE, $group_id);
      $unpublished_nodes = $this->getNodeCount($content_type->id(), FALSE, $group_id);

      $content_type_data['content_type_id'] = $type_id;
      $content_type_data['content_type_name'] = $type_name;
      $content_type_data['content_type_type'] = $type;
      $content_type_data['content_type_description'] = $type_description;

      switch ($percentage_totals) {
        case "published_nodes":
          // Sum of all published nodes.
          $count_published_nodes = $this->getNodeCountAll(TRUE, $group_id);
          $content_type_data['published_percent'] = $count_published_nodes ? ($published_nodes / $count_published_nodes) * 100 : 0;
          $content_type_data['percentage_totals'] = "* Published nodes";
          break;

        case "all_nodes":
          // Sum of all nodes.
          $count_published_nodes = $this->getAllNodeCount($group_id);
          $content_type_data['published_percent'] = $count_published_nodes ? (($published_nodes + $unpublished_nodes) / $this->getAllNodeCount($group_id)) * 100 : 0;
          $content_type_data['percentage_totals'] = "* All nodes";
          break;

        case "unpublished_nodes":
          // Sum of all published nodes.
          $count_unpublished_nodes = $this->getNodeCountAll(FALSE, $group_id);
          $content_type_data['published_percent'] = $count_unpublished_nodes ? ($unpublished_nodes / $count_unpublished_nodes) * 100 : 0;
          $content_type_data['percentage_totals'] = "* Unpublished nodes";
          break;

      }

      $content_type_data['total_count'] = $published_nodes + $unpublished_nodes;
      $content_type_data['published_count'] = $published_nodes;
      $content_type_data['unpublished_count'] = $unpublished_nodes;

      // Fetch insights for # months.
      $content_type_data['insights'] = $this->getInsightsCounts($content_type->id(), $months_ago, $current_date, $group_id);

      // Moderation State Report.
      if ($this->moduleHandler->moduleExists('content_moderation')) {
        $content_type_data['moderation_state'] = $this->getNodeSummaryPerModerationState($content_type->id(), $moderation_state, $group_id);
      }

      // Add data to report.
      $report_data[] = $content_type_data;
    }

    // Sort result by.
    switch ($sort_by_field) {
      case "content_type":
        switch ($sort_by_direction) {
          case "asc":
            usort($report_data, function ($a, $b) {
              return strcmp($a["content_type_name"], $b["content_type_name"]);
            });
            break;

          case "desc":
            usort($report_data, function ($a, $b) {
              return strcmp($b["content_type_name"], $a["content_type_name"]);
            });
            break;

        }
        break;

      case "content_type_machine":
        switch ($sort_by_direction) {
          case "asc":
            usort($report_data, function ($a, $b) {
              return strcmp($a["content_type_type"], $b["content_type_type"]);
            });
            break;

          case "desc":
            usort($report_data, function ($a, $b) {
              return strcmp($b["content_type_type"], $a["content_type_type"]);
            });
            break;

        }
        break;

      case "total_percentage":
        switch ($sort_by_direction) {
          case "asc":
            usort($report_data, function ($a, $b) {
              return ($a['published_percent'] - $b['published_percent']) ? ($a['published_percent'] - $b['published_percent']) / abs($a['published_percent'] - $b['published_percent']) : 0;
            });
            break;

          case "desc":
            usort($report_data, function ($a, $b) {
              return ($b['published_percent'] - $a['published_percent']) ? ($b['published_percent'] - $a['published_percent']) / abs($b['published_percent'] - $a['published_percent']) : 0;
            });
            break;

        }
        break;

    }

    $total_nodes = [];
    $total_nodes['total_nodes'] = $this->getAllNodeCount($group_id);
    $total_nodes['total_nodes_published'] = $this->getNodeCountAll(TRUE, $group_id);
    $total_nodes['total_nodes_unpublished'] = $this->getNodeCountAll(FALSE, $group_id);

    // Get current path for img.
    $modulePath = $this->moduleHandler->getModule('content_insights_report_group')->getPath();
    return [
      '#group' => $group_data,
      '#theme' => 'content_insights_report_group',
      '#show_summary' => $show_summary,
      '#show_insights_report' => $show_insights_report,
      '#show_workflow_moderation_state_report' => $show_workflow_moderation_state_report,
      '#show_created_by_and_date' => $show_created_by_and_date,
      '#show_empty_results' => $show_empty_results,
      '#months' => $number_of_months,
      '#show_print' => $show_print,
      '#print_icon' => $modulePath . "/images/print-icon.png",
      '#report_data' => $report_data,
      '#total_nodes' => $total_nodes,
      '#percentage_totals' => $percentage_totals_text,
      '#attached' => [
        'library' => [
          'content_insights_report_group/content_insights_report_group',
        ],
      ],
    ];
  }

  /**
   * Count all number of nodes.
   *
   * @return int
   *   Count value of number of nodes.
   */
  private function getAllNodeCount($group_id) {

    // Get the database connection.
    $database = Database::getConnection();

    // Query the group_relationship_field_data table to find nodes associated
    // with the group.
    $query = $database->select('node_field_data', 'nfd');
    $query->fields('nfd', ['nid', 'status']);
    $query->innerJoin('group_relationship_field_data', 'grfd', "nfd.nid = grfd.entity_id");
    $query->condition('grfd.gid', $group_id);
    $query->condition('grfd.plugin_id', "group_node:%", 'LIKE');

    return count($query->execute()->fetchAll());
  }

  /**
   * Count number of nodes based on status.
   *
   * @return int
   *   Count value of number of nodes.
   */
  private function getNodeCountAll($is_published, $group_id) {
    // Get the database connection.
    $database = Database::getConnection();

    // Query the group_relationship_field_data table to find nodes associated
    // with the group.
    $query = $database->select('node_field_data', 'nfd');
    $query->fields('nfd', ['nid', 'status']);
    $query->innerJoin('group_relationship_field_data', 'grfd', "nfd.nid = grfd.entity_id");
    $query->condition('grfd.gid', $group_id);
    $query->condition('grfd.plugin_id', "group_node:%", 'LIKE');
    $query->condition('nfd.status', $is_published ? 1 : 0);

    return count($query->execute()->fetchAll());
  }

  /**
   * Count number of nodes based on status.
   *
   * @return int
   *   Count value of number of nodes.
   */
  private function getNodeCount($content_type_id, $is_published, $group_id) {

    // Get the database connection.
    $database = Database::getConnection();

    // Query the group_relationship_field_data table to find nodes associated
    // with the group.
    $query = $database->select('node_field_data', 'nfd');
    $query->fields('nfd', ['nid', 'status', 'type']);
    $query->condition('nfd.status', $is_published ? 1 : 0);
    $query->condition('nfd.type', $content_type_id);
    $query->innerJoin('group_relationship_field_data', 'grfd', "nfd.nid = grfd.entity_id");
    $query->condition('grfd.gid', $group_id);
    $query->condition('grfd.plugin_id', "group_node:%", 'LIKE');

    return count($query->execute()->fetchAll());
  }

  /**
   * Insights Counts values.
   *
   * @return array
   *   An array Insights.
   */
  private function getInsightsCounts($content_type_id, $start_date, $end_date, $group_id) {
    $insights_counts = [];

    $config = $this->configFactory->get('content_insights_report_group.settings');
    $number_of_months = $config->get('number_of_months');
    $month_sort_by_direction = $config->get('month_sort_by_direction');

    for ($i = 1; $i <= $number_of_months; $i++) {
      $start_of_month = clone $start_date;
      $start_of_month = $start_of_month->modify('first day of ' . ($i === 0 ? 'this month' : "+$i months"));
      $end_of_month = clone $start_of_month;
      $end_of_month->modify('last day of this month');

      $insights_counts[] = [
        'year_month' => (int) $start_of_month->format('Ym'),
        'month' => $start_of_month->format('F Y'),
        'created_count' => $this->getNodeCreatedCount($content_type_id, $start_of_month, $end_of_month, $group_id),
        'updated_count' => $this->getNodeUpdatedCount($content_type_id, $start_of_month, $end_of_month, $group_id),
      ];
    }

    // Sort Month by.
    switch ($month_sort_by_direction) {
      case "asc":
        usort($insights_counts, function ($a, $b) {
          return strcmp($a['year_month'], $b['year_month']);
        });
        break;

      case "desc":
        usort($insights_counts, function ($a, $b) {
          return strcmp($b['year_month'], $a['year_month']);
        });
        break;

    }

    return $this->flipRowColArray($insights_counts);
  }

  /**
   * Flip array row to col.
   *
   * @return array
   *   An array of flipped values.
   */
  private function flipRowColArray($array) {
    $out = [];
    foreach ($array as $rowkey => $row) {
      foreach ($row as $colkey => $col) {
        $out[$colkey][$rowkey] = $col;
      }
    }
    return $out;
  }

  /**
   * Count number of nodes based on dates created.
   *
   * @return int
   *   Count value of number of nodes.
   */
  private function getNodeCreatedCount($content_type_id, $start_date, $end_date, $group_id) {

    // Get the database connection.
    $database = Database::getConnection();

    // Query the group_relationship_field_data table to find nodes associated
    // with the group.
    $query = $database->select('node_field_data', 'nfd');
    $query->fields('nfd', ['nid', 'status', 'type']);
    $query->condition('nfd.created',
      [
        $start_date->getTimestamp(),
        $end_date->getTimestamp(),
      ],
      'BETWEEN');
    $query->condition('nfd.type', $content_type_id);
    $query->innerJoin('group_relationship_field_data', 'grfd', "nfd.nid = grfd.entity_id");
    $query->condition('grfd.gid', $group_id);
    $query->condition('grfd.plugin_id', "group_node:%", 'LIKE');

    return count($query->execute()->fetchAll());
  }

  /**
   * Count number of nodes based on dates updated.
   *
   * @return int
   *   Count value of number of nodes.
   */
  private function getNodeUpdatedCount($content_type_id, $start_date, $end_date, $group_id) {

    // Get the database connection.
    $database = Database::getConnection();

    // Query the group_relationship_field_data table to find nodes associated
    // with the group.
    $query = $database->select('node_field_data', 'nfd');
    $query->fields('nfd', ['nid', 'status', 'type']);
    $query->condition('nfd.changed',
      [
        $start_date->getTimestamp(),
        $end_date->getTimestamp(),
      ],
      'BETWEEN');
    $query->condition('nfd.type', $content_type_id);
    $query->innerJoin('group_relationship_field_data', 'grfd', "nfd.nid = grfd.entity_id");
    $query->condition('grfd.gid', $group_id);
    $query->condition('grfd.plugin_id', "group_node:%", 'LIKE');

    return count($query->execute()->fetchAll());
  }

  /**
   * Generates based on content type, status, and moderation state.
   *
   * @return array
   *   An array of counts grouped by content type, status, and moderation state.
   */
  private function getNodeSummaryPerModerationState($content_type_id, $moderation_state, $group_id) {

    $connection = Database::getConnection();

    // Query to count nodes grouped by content type and moderation state.
    $query = $connection->select('node_field_data', 'nfd');
    $query->fields('nfd', ['type', 'status']);
    $query->addField('cmsfd', 'workflow', 'workflow');
    $query->addField('cmsfd', 'moderation_state', 'moderation_state');
    $query->addExpression('COUNT(nfd.nid)', 'node_count');
    $query->leftJoin('content_moderation_state_field_data', 'cmsfd', 'nfd.nid = cmsfd.content_entity_id');
    $query->condition('nfd.type', $content_type_id);
    $query->innerJoin('group_relationship_field_data', 'grfd', "nfd.nid = grfd.entity_id");
    $query->condition('grfd.gid', $group_id);
    $query->condition('grfd.plugin_id', "group_node:%", 'LIKE');
    // Don't remove (website will load long and time out).
    $query->isNotNull('cmsfd.moderation_state');
    $query->groupBy('nfd.type');
    $query->groupBy('nfd.status');
    $query->groupBy('cmsfd.workflow');
    $query->groupBy('cmsfd.moderation_state');

    $result = $query->execute();

    foreach ($result as $record) {
      $moderation_state[$record->workflow]['status'][$record->status]['stage'][$record->moderation_state ?: 'no_moderation_state']['count'] = $record->node_count;
      $moderation_state[$record->workflow]['total_count']++;
    }

    return $moderation_state;
  }

}
